package mathvis.halfedge;

import de.jtem.halfedge.Edge;
import de.jtem.halfedge.Face;
import de.jtem.halfedge.Vertex;

public class HEDSUtil {

	/**
	 * Test whether the half-edge data structure represents a valid surface and, optionally, give a reason if it fails the test.
	 * If the parameter is {@code false}, this method behaves exatly as {@link #isValidSurface(HalfEdgeDataStructure)}. If the parameter is {@code true},
	 * and the half-edge data structure fails to represent a valid surface, the method outputs a brief explanation to {@link System#err}.
	 * This feature is intended for debugging.
	 * @param printReasonForFailureToSystemErr {@code true} if you want output to {@link System#err}.
	 * @return {@code true} if the half-edge data structure represents a valid surface, {@code false} otherwise
	 */
	static public boolean isValidSurface(HEDS heds) {
		
		boolean  printReasonForFailureToSystemErr = true;

		// must have at least one edge
		if (heds.getEdges().isEmpty()) {
			if (printReasonForFailureToSystemErr) System.err.println("getEdges().isEmpty == true");
			return false;
		}

		// check for null references
		for (Edge<?,?,?> e : heds.getEdges()) {
			if (e.getNextEdge() == null) {
				if (printReasonForFailureToSystemErr) System.err.println("getNextEdge() returns null on " + e);
				return false;
			}
			if (e.getPreviousEdge() == null) {
				if (printReasonForFailureToSystemErr) System.err.println("getPreviousEdge() returns null on " + e);
				return false;
			}
			if (e.getOppositeEdge() == null) {
				if (printReasonForFailureToSystemErr) System.err.println("getOppositeEdge() returns null on " + e);
				return false;
			}
			if (e.getTargetVertex() == null) {
				if (printReasonForFailureToSystemErr) System.err.println("getTargetVertex() returns null on " + e);
				return false;
			}
			// either left face or right face may be null but not both
			if (e.getLeftFace() == null && e.getRightFace() == null) {
				if (printReasonForFailureToSystemErr) System.err.println("Left face and right face are null for edge " + e + ".");
				return false;
			}
		}

		// Check if edges in a cycle have same left face and edges in a cocycle have same target vertex
		for (Edge<?,?,?> e : heds.getEdges()) {
			if (e.getLeftFace() != e.getNextEdge().getLeftFace()) {
				if (printReasonForFailureToSystemErr) System.err.println("e.getLeftFace() != e.getNextEdge().getLeftFace() for e=" + e);
				return false;
			}
			if (e.getTargetVertex() != e.getNextEdge().getOppositeEdge().getTargetVertex()) {
				if (printReasonForFailureToSystemErr) System.err.println("e.getTargetVertex() != e.getNextEdge().getOppositeEdge().getTargetVertex() for e=" + e);
				return false;
			}
		}		

		// check if every face has a boundary edge
		for (Face<?,?,?> f : heds.getFaces()) {
			if (f.getBoundaryEdge() == null) {
				if (printReasonForFailureToSystemErr) System.err.println(f + " has no boundary edge.");
				return false;
			}
		}
		
		// check if every vertex has an incoming edge
		for (Vertex<?,?,?> v : heds.getVertices()) {
			if (v.getIncomingEdge() == null) {
				if (printReasonForFailureToSystemErr) System.err.println(v + " has no incoming vertex.");
			}
		}
		
		// check if each face corresponds to a unique edge cycle, 
		// each vertex corresponds to a unique edge cocycle,
		// and that there is at most one edge in a vertex cocyle with left face == null 
		int ne = heds.numEdges();
		int nf = heds.numFaces();
		int nv = heds.numVertices();
		if (!(nv > 0 && ne > 0 && nf > 0)) {
			throw new RuntimeException("Test failed: nv="+nv+" ne="+ne+" nv="+nv);
		}
		boolean[] vertexMark = new boolean[nv];
		boolean[] edgeMark = new boolean[ne];
		boolean[] faceMark = new boolean[nf];
		for (Edge<?,?,?> e : heds.getEdges()) {
			if (edgeMark[e.getIndex()]) {
				// the cycle of this edge has already been treated
				continue;
			}
			Face<?,?,?> f = e.getLeftFace();
			if (f != null) {
				// has this edge's left face already occurred in another cycle?
				if (faceMark[f.getIndex()]) {
					if (printReasonForFailureToSystemErr) System.err.println(f + " is contained in more than one edge cycle. (Found out while looking at " + e + ")");
					return false;
				}
				// mark the left face
				faceMark[f.getIndex()] = true;
			}
			// mark all edges in the edge cycle
			edgeMark[e.getIndex()] = true;
			for (Edge<?,?,?> e1 = e.getNextEdge(); e1 != e; e1 = e1.getNextEdge()) {
//				System.err.println(e1);
				edgeMark[e1.getIndex()] = true;
			}
		}
		edgeMark = new boolean[ne]; // initialize to false again
		for (Edge<?,?,?> e : heds.getEdges()) {
			if (edgeMark[e.getIndex()]) {
				// the cycle of this edge has already been treated
				continue;
			}
			Vertex<?,?,?> v = e.getTargetVertex();
			// has this edge's target vertex already occurred in another cycle?
			if (vertexMark[v.getIndex()]) {
				if (printReasonForFailureToSystemErr) System.err.println(v + " is contained in more than one edge cocycle.");
				return false;
			}
			// mark the left face
			vertexMark[v.getIndex()] = true;
			// mark all edges in the edge cocycle and check whether there's more than one with left face == null.
			edgeMark[e.getIndex()] = true;
			boolean leftFaceNull = e.getLeftFace() == null;
			for (Edge<?,?,?> e1 = e.getNextEdge().getOppositeEdge(); e1 != e; e1 = e1.getNextEdge().getOppositeEdge()) {
//				System.err.println(e1);
				edgeMark[e1.getIndex()] = true;
				if (e1.getLeftFace() == null) {
					if (leftFaceNull) {
						if (printReasonForFailureToSystemErr) System.err.println("There is more than one edge with target " + v + " and left face null.");
						return false;
					}
					leftFaceNull = true;
				}
			}
		}

		// passed all tests
		return true;
	}

}
